@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
- * Copyright (c) [2016-2018] SUSE LLC
+ * Copyright (c) [2016-2024] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -23,10 +23,10 @@
 
 #include "config.h"
 
-#include <string.h>
+#include <cstring>
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <errno.h>
+#include <cerrno>
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
@@ -44,18 +44,19 @@
 #include <boost/version.hpp>
 #include <boost/thread.hpp>
 #endif
+#include <regex>
 #include <boost/algorithm/string.hpp>
 
 #include "snapper/Log.h"
 #include "snapper/Btrfs.h"
 #include "snapper/BtrfsUtils.h"
 #include "snapper/File.h"
+#include "snapper/PluginsImpl.h"
 #include "snapper/Snapper.h"
 #include "snapper/SnapperTmpl.h"
 #include "snapper/SnapperDefines.h"
 #include "snapper/Acls.h"
 #include "snapper/Exception.h"
-#include "snapper/Regex.h"
 #ifdef ENABLE_ROLLBACK
 #include "snapper/MntTable.h"
 #endif
@@ -75,7 +76,7 @@ namespace snapper
 	if (fstype == "btrfs")
 	    return new Btrfs(subvolume, root_prefix);
 
-	return NULL;
+	return nullptr;
     }
 
 
@@ -91,7 +92,7 @@ namespace snapper
 #ifdef ENABLE_BTRFS_QUOTA
 
 	string qgroup_str;
-	if (config_info.getValue("QGROUP", qgroup_str) && !qgroup_str.empty())
+	if (config_info.get_value("QGROUP", qgroup_str) && !qgroup_str.empty())
 	{
 	    try
 	    {
@@ -102,6 +103,19 @@ namespace snapper
 		y2err("failed to parse qgroup '" << qgroup_str << "'");
 		SN_THROW(InvalidConfigException());
 	    }
+
+	    if (get_level(qgroup) == 0)
+	    {
+		y2err("invalid level of qgroup '" << qgroup_str << "'");
+		SN_THROW(InvalidConfigException());
+	    }
+
+	    SDir general_dir = openGeneralDir();
+	    if (!does_qgroup_exist(general_dir.fd(), qgroup))
+	    {
+		y2err("qgroup '" << qgroup_str << "' does not exist");
+		qgroup = no_qgroup;
+	    }
 	}
 	else
 	{
@@ -119,7 +133,7 @@ namespace snapper
 
 	try
 	{
-	    create_subvolume(subvolume_dir.fd(), ".snapshots");
+	    create_subvolume(subvolume_dir.fd(), SNAPSHOTS_NAME);
 	}
 	catch (const runtime_error_with_errno& e)
 	{
@@ -130,13 +144,14 @@ namespace snapper
 		case EEXIST:
 		    SN_THROW(CreateConfigFailedException("creating btrfs subvolume .snapshots failed "
 							 "since it already exists"));
+		    break;
 
 		default:
 		    SN_THROW(CreateConfigFailedException("creating btrfs subvolume .snapshots failed"));
 	    }
 	}
 
-	SFile x(subvolume_dir, ".snapshots");
+	SFile x(subvolume_dir, SNAPSHOTS_NAME);
 #ifdef ENABLE_SELINUX
 	try
 	{
@@ -164,7 +179,7 @@ namespace snapper
 #ifdef ENABLE_ROLLBACK
 	if (subvolume == "/")
 	{
-	    subvolume_dir.umount(".snapshots");
+	    subvolume_dir.umount(SNAPSHOTS_NAME);
 
 	    removeFromFstab();
 	}
@@ -172,7 +187,7 @@ namespace snapper
 
 	try
 	{
-	    delete_subvolume(subvolume_dir.fd(), ".snapshots");
+	    delete_subvolume(subvolume_dir.fd(), SNAPSHOTS_NAME);
 	}
 	catch (const runtime_error& e)
 	{
@@ -223,8 +238,8 @@ namespace snapper
     string
     Btrfs::snapshotDir(unsigned int num) const
     {
-	return (subvolume == "/" ? "" : subvolume) + "/.snapshots/" + decString(num) +
-	    "/snapshot";
+	return (subvolume == "/" ? "" : subvolume) + "/" SNAPSHOTS_NAME "/" + decString(num) +
+	    "/" SNAPSHOT_NAME;
     }
 
 
@@ -252,7 +267,7 @@ namespace snapper
     Btrfs::openInfosDir() const
     {
 	SDir subvolume_dir = openSubvolumeDir();
-	SDir infos_dir(subvolume_dir, ".snapshots");
+	SDir infos_dir(subvolume_dir, SNAPSHOTS_NAME);
 
 	struct stat stat;
 	if (infos_dir.stat(&stat) != 0)
@@ -291,7 +306,7 @@ namespace snapper
     Btrfs::openSnapshotDir(unsigned int num) const
     {
 	SDir info_dir = openInfoDir(num);
-	SDir snapshot_dir(info_dir, "snapshot");
+	SDir snapshot_dir(info_dir, SNAPSHOT_NAME);
 
 	return snapshot_dir;
     }
@@ -316,9 +331,9 @@ namespace snapper
 	    try
 	    {
 		if (empty)
-		    create_subvolume(info_dir.fd(), "snapshot");
+		    create_subvolume(info_dir.fd(), SNAPSHOT_NAME);
 		else
-		    create_snapshot(subvolume_dir.fd(), info_dir.fd(), "snapshot", read_only,
+		    create_snapshot(subvolume_dir.fd(), info_dir.fd(), SNAPSHOT_NAME, read_only,
 				    quota ? qgroup : no_qgroup);
 	    }
 	    catch (const runtime_error& e)
@@ -334,7 +349,7 @@ namespace snapper
 
 	    try
 	    {
-		create_snapshot(snapshot_dir.fd(), info_dir.fd(), "snapshot", read_only,
+		create_snapshot(snapshot_dir.fd(), info_dir.fd(), SNAPSHOT_NAME, read_only,
 				quota ? qgroup : no_qgroup);
 	    }
 	    catch (const runtime_error& e)
@@ -372,7 +387,7 @@ namespace snapper
 
 	try
 	{
-	    create_snapshot(tmp_mount_dir.fd(), info_dir.fd(), "snapshot", read_only,
+	    create_snapshot(tmp_mount_dir.fd(), info_dir.fd(), SNAPSHOT_NAME, read_only,
 			    quota ? qgroup : no_qgroup);
 	}
 	catch (const runtime_error& e)
@@ -400,38 +415,19 @@ namespace snapper
 
 	try
 	{
-#ifdef HAVE_LIBBTRFS
+#if defined(HAVE_LIBBTRFS) || defined(HAVE_LIBBTRFSUTIL)
 	    subvolid_t subvolid = get_id(openSnapshotDir(num).fd());
 #endif
 
-	    delete_subvolume(info_dir.fd(), "snapshot");
+	    delete_subvolume(info_dir.fd(), SNAPSHOT_NAME);
 
-#ifdef HAVE_LIBBTRFS
+#if defined(HAVE_LIBBTRFS) || defined(HAVE_LIBBTRFSUTIL)
 	    deleted_subvolids.push_back(subvolid);
-#endif
-
-#ifdef ENABLE_BTRFS_QUOTA
-
-	    // workaround for the kernel not deleting the qgroup of a
-	    // subvolume when deleting the subvolume, see
-	    // https://bugzilla.suse.com/show_bug.cgi?id=972511
-
-	    try
-	    {
-		SDir subvolume_dir = openSubvolumeDir();
-		qgroup_destroy(subvolume_dir.fd(), calc_qgroup(0, subvolid));
-	    }
-	    catch (const runtime_error& e)
-	    {
-		// Ignore that the qgroup could not be destroyed. Should not
-		// cause problems except of having unused qgroups.
-	    }
-
 #endif
 	}
 	catch (const runtime_error& e)
 	{
-	    y2err("delete snapshot failed, " << e.what());
+	    y2err("delete snapshot " << info_dir.fullname() << "/" SNAPSHOT_NAME " failed, " << e.what());
 	    SN_THROW(DeleteSnapshotFailedException());
 	}
     }
@@ -464,6 +460,14 @@ namespace snapper
     }
 
 
+    void
+    Btrfs::setSnapshotReadOnly(unsigned int num, bool read_only) const
+    {
+	SDir snapshot_dir = openSnapshotDir(num);
+	set_subvolume_read_only(snapshot_dir.fd(), read_only);
+    }
+
+
     bool
     Btrfs::checkSnapshot(unsigned int num) const
     {
@@ -472,13 +476,16 @@ namespace snapper
 	    SDir info_dir = openInfoDir(num);
 
 	    struct stat stat;
-	    int r = info_dir.stat("snapshot", &stat, AT_SYMLINK_NOFOLLOW);
+	    int r = info_dir.stat(SNAPSHOT_NAME, &stat, AT_SYMLINK_NOFOLLOW);
 	    return r == 0 && is_subvolume(stat);
 	}
 	catch (const IOErrorException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    // TODO the openInfoDir above logs an error although when this
 	    // function is used from nextNumber the failure is ok
+
 	    return false;
 	}
     }
@@ -501,7 +508,7 @@ namespace snapper
 
 	int status;
 
-	map<string, tree_node> childs;
+	map<string, tree_node> children;
 
 	tree_node* find(const string& name);
 
@@ -527,8 +534,8 @@ namespace snapper
 	string::size_type pos = name.find('/');
 	if (pos == string::npos)
 	{
-	    iterator it = childs.find(name);
-	    if (it == childs.end())
+	    iterator it = children.find(name);
+	    if (it == children.end())
 		return NULL;
 
 	    return &it->second;
@@ -536,8 +543,8 @@ namespace snapper
 	else
 	{
 	    string a = name.substr(0, pos);
-	    iterator it = childs.find(a);
-	    if (it == childs.end())
+	    iterator it = children.find(a);
+	    if (it == children.end())
 		return NULL;
 
 	    string b = name.substr(pos + 1);
@@ -552,18 +559,18 @@ namespace snapper
 	string::size_type pos = name.find('/');
 	if (pos == string::npos)
 	{
-	    iterator it = childs.find(name);
-	    if (it == childs.end())
-		it = childs.insert(childs.end(), make_pair(name, tree_node()));
+	    iterator it = children.find(name);
+	    if (it == children.end())
+		it = children.insert(children.end(), make_pair(name, tree_node()));
 
 	    return &it->second;
 	}
 	else
 	{
 	    string a = name.substr(0, pos);
-	    iterator it = childs.find(a);
-	    if (it == childs.end())
-		it = childs.insert(childs.end(), make_pair(a, tree_node()));
+	    iterator it = children.find(a);
+	    if (it == children.end())
+		it = children.insert(children.end(), make_pair(a, tree_node()));
 
 	    string b = name.substr(pos + 1);
 	    return it->second.insert(b);
@@ -577,12 +584,12 @@ namespace snapper
 	string::size_type pos = name.find('/');
 	if (pos == string::npos)
 	{
-	    iterator it = childs.find(name);
-	    if (it == childs.end())
+	    iterator it = children.find(name);
+	    if (it == children.end())
 		return false;
 
-	    if (it->second.childs.empty())
-		childs.erase(it);
+	    if (it->second.children.empty())
+		children.erase(it);
 	    else
 		it->second.status = 0;
 
@@ -591,15 +598,15 @@ namespace snapper
 	else
 	{
 	    string a = name.substr(0, pos);
-	    iterator it = childs.find(a);
-	    if (it == childs.end())
+	    iterator it = children.find(a);
+	    if (it == children.end())
 		return false;
 
 	    string b = name.substr(pos + 1);
 	    it->second.erase(b);
 
-	    if (it->second.status == 0 && it->second.childs.empty())
-		childs.erase(it);
+	    if (it->second.status == 0 && it->second.children.empty())
+		children.erase(it);
 
 	    return true;
 	}
@@ -618,7 +625,7 @@ namespace snapper
 	    return false;
 
 	nn = insert(n);
-	swap(nn->childs, oo->childs);
+	swap(nn->children, oo->children);
 	nn->status = oo->status;
 	erase(o);
 
@@ -629,7 +636,7 @@ namespace snapper
     void
     tree_node::dump(const string& prefix) const
     {
-	for (const_iterator it = childs.begin(); it != childs.end(); ++it)
+	for (const_iterator it = children.begin(); it != children.end(); ++it)
 	{
 	    if (prefix.empty())
 	    {
@@ -676,10 +683,6 @@ namespace snapper
 
 	bool dumper(int fd);
 
-#if BOOST_VERSION < 104100
-	bool dumper_ret;
-#endif
-
 	void do_send(u64 parent_root_id, const vector<u64>& clone_sources);
 
     };
@@ -712,7 +715,7 @@ namespace snapper
     void
     tree_node::check(StreamProcessor* processor, const string& prefix)
     {
-	for (iterator it = childs.begin(); it != childs.end(); ++it)
+	for (iterator it = children.begin(); it != children.end(); ++it)
 	{
 	    if (prefix.empty())
 	    {
@@ -731,7 +734,7 @@ namespace snapper
     void
     tree_node::result(cmpdirs_cb_t cb, const string& prefix) const
     {
-	for (const_iterator it = childs.begin(); it != childs.end(); ++it)
+	for (const_iterator it = children.begin(); it != children.end(); ++it)
 	{
 	    if (prefix.empty())
 	    {
@@ -905,7 +908,7 @@ namespace snapper
     merge(StreamProcessor* processor, tree_node* tmp, const string& from, const string& to,
 	  const string& prefix = "")
     {
-	for (tree_node::iterator it = tmp->childs.begin(); it != tmp->childs.end(); ++it)
+	for (tree_node::iterator it = tmp->children.begin(); it != tmp->children.end(); ++it)
 	{
 	    if (prefix.empty())
 	    {
@@ -992,7 +995,7 @@ namespace snapper
 	    else
 	    {
 		tree_node tmp;
-		swap(it1->childs, tmp.childs);
+		swap(it1->children, tmp.children);
 
 		processor->deleted(from);
 		processor->created(to);
@@ -1235,11 +1238,15 @@ namespace snapper
     bool
     StreamProcessor::dumper(int fd)
     {
+	FdCloser fd_closer(fd);
+
+	unsigned int iterations = 0;
+
 	while (true)
 	{
 	    boost::this_thread::interruption_point();
 
-	     // remove the fourth parameter for older versions of libbtrfs
+	    // remove the fourth parameter for older versions of libbtrfs
 	    int r;
 
 #if BTRFS_LIB_VERSION < 101
@@ -1248,26 +1255,25 @@ namespace snapper
 	    r = btrfs_read_and_process_send_stream(fd, &send_ops, &*this, 0, 1);
 #endif
 
-	    if (r < 0)
-	    {
-		y2err("btrfs_read_and_process_send_stream failed");
+	    // Only return an error when r == -ENODATA if this was the
+	    // first call to btrfs_read_and_process_send_stream(). See
+	    // https://github.com/openSUSE/snapper/pull/438,
+	    // https://bugzilla.suse.com/show_bug.cgi?id=1111414 and
+	    // the btrfs-progs source code.
 
-#if BOOST_VERSION < 104100
-		dumper_ret = false;
-#endif
+	    if (r < 0 && !(r == -ENODATA && iterations > 0))
+	    {
+		y2err("btrfs_read_and_process_send_stream failed " << r);
 
 		return false;
 	    }
 
 	    if (r)
 	    {
-#if BOOST_VERSION < 104100
-		dumper_ret = true;
-#endif
-
 		return true;
 	    }
 
+	    ++iterations;
 	}
     }
 
@@ -1283,6 +1289,10 @@ namespace snapper
 	    SN_THROW(BtrfsSendReceiveException());
 	}
 
+	// Use RAII to help close fds.
+	FdCloser fd0_closer(pipefd[0]);
+	FdCloser fd1_closer(pipefd[1]);
+
 	struct btrfs_ioctl_send_args io_send;
 	memset(&io_send, 0, sizeof(io_send));
 	io_send.send_fd = pipefd[1];
@@ -1291,65 +1301,27 @@ namespace snapper
 	io_send.parent_root = parent_root_id;
 	io_send.flags = BTRFS_SEND_FLAG_NO_FILE_DATA;
 
-#if BOOST_VERSION >= 104100
-
 	boost::packaged_task<bool> pt(boost::bind(&StreamProcessor::dumper, this, pipefd[0]));
 	boost::unique_future<bool> uf = pt.get_future();
 
 	boost::thread task(boost::move(pt));
 
+	fd0_closer.reset();
+
 	int r2 = ioctl(dir2.fd(), BTRFS_IOC_SEND, &io_send);
 	if (r2 < 0)
 	{
 	    y2err("send ioctl failed errno:" << errno << " (" << stringerror(errno) << ")");
 	}
 
-	close(pipefd[1]);
+	fd1_closer.close();
 
 	uf.wait();
 
-	close(pipefd[0]);
-
 	if (r2 < 0 || !uf.get())
 	{
 	    SN_THROW(BtrfsSendReceiveException());
 	}
-
-#else
-
-	boost::thread dumper_thread(boost::bind(&StreamProcessor::dumper, this, pipefd[0]));
-
-	int r2 = ioctl(dir2.fd(), BTRFS_IOC_SEND, &io_send);
-	if (r2 < 0)
-	{
-	    y2err("send ioctl failed errno:" << errno << " (" << stringerror(errno) << ")");
-	}
-
-	close(pipefd[1]);
-
-	dumper_thread.join();
-
-	close(pipefd[0]);
-
-	if (r2 < 0 || !dumper_ret)
-	{
-	    SN_THROW(BtrfsSendReceiveException());
-	}
-
-#endif
-    }
-
-
-    static bool
-    is_subvolume_ro(const SDir& dir)
-    {
-	u64 flags;
-	if (ioctl(dir.fd(), BTRFS_IOC_SUBVOL_GETFLAGS, &flags) < 0)
-	{
-	    SN_THROW(IOErrorException("ioctl BTRFS_IOC_SUBVOL_GETFLAGS failed"));
-	}
-
-	return flags & BTRFS_SUBVOL_RDONLY;
     }
 
 
@@ -1358,7 +1330,7 @@ namespace snapper
     {
 	y2mil("dir1:'" << dir1.fullname() << "' dir2:'" << dir2.fullname() << "'");
 
-	if (!is_subvolume_ro(dir1) || !is_subvolume_ro(dir2))
+	if (!is_subvolume_read_only(dir1.fd()) || !is_subvolume_read_only(dir2.fd()))
 	{
 	    y2err("not read-only snapshots");
 	    SN_THROW(BtrfsSendReceiveException());
@@ -1389,7 +1361,7 @@ namespace snapper
 
 	try
 	{
-	    StopWatch stopwatch;
+	    Stopwatch stopwatch;
 
 	    const SDir subvolume(openSubvolumeDir());
 
@@ -1453,19 +1425,20 @@ namespace snapper
 
 
     std::pair<bool, unsigned int>
-    Btrfs::getDefault() const
+    Btrfs::idToNum(int fd, subvolid_t id) const
     {
-	SDir subvolume_dir = openSubvolumeDir();
+	string path = get_subvolume(fd, id);
 
-	subvolid_t id = get_default_id(subvolume_dir.fd());
+	static const regex rx("/([0-9]+)/snapshot$", regex::extended);
+	smatch match;
 
-	string path = get_subvolume(subvolume_dir.fd(), id);
-
-	Regex rx("/([0-9]+)/snapshot$");
-	if (!rx.match(path))
+	if (!regex_search(path, match, rx))
 	    return make_pair(false, 0);
 
-	unsigned int num = stoi(rx.cap(1));
+	unsigned int num = stoi(match[1].str());
+
+	if (!checkSnapshot(num))
+	    return make_pair(false, 0);
 
 	if (get_id(openSnapshotDir(num).fd()) != id)
 	    return make_pair(false, 0);
@@ -1474,11 +1447,39 @@ namespace snapper
     }
 
 
-    void
-    Btrfs::setDefault(unsigned int num) const
+    std::pair<bool, unsigned int>
+    Btrfs::getDefault() const
     {
 	try
 	{
+	    SDir subvolume_dir = openSubvolumeDir();
+	    int fd = subvolume_dir.fd();
+
+	    return idToNum(fd, get_default_id(fd));
+	}
+	catch (const exception& e)
+	{
+	    SN_THROW(IOErrorException(string("query default id failed, ") + e.what()));
+	    __builtin_unreachable();
+	}
+    }
+
+
+    void
+    Btrfs::setDefault(unsigned int num) const
+    {
+	Plugins::Report report;
+	setDefault(num, report);
+    }
+
+
+    void
+    Btrfs::setDefault(unsigned int num, Plugins::Report& report) const
+    {
+	try
+	{
+	    Plugins::set_default_snapshot(Plugins::Stage::PRE_ACTION, subvolume, this, num, report);
+
 	    SDir general_dir = openGeneralDir();
 
 	    if (num == 0)
@@ -1493,6 +1494,8 @@ namespace snapper
 		subvolid_t id = get_id(snapshot_dir.fd());
 		set_default_id(general_dir.fd(), id);
 	    }
+
+	    Plugins::set_default_snapshot(Plugins::Stage::POST_ACTION, subvolume, this, num, report);
 	}
 	catch (const runtime_error& e)
 	{
@@ -1501,6 +1504,16 @@ namespace snapper
     }
 
 
+    std::pair<bool, unsigned int>
+    Btrfs::getActive() const
+    {
+	SDir subvolume_dir = openSubvolumeDir();
+	int fd = subvolume_dir.fd();
+
+	return idToNum(fd, get_id(fd));
+    }
+
+
     bool
     Btrfs::isActive(unsigned int num) const
     {
@@ -1528,28 +1541,42 @@ namespace snapper
     bool
     Btrfs::isDefault(unsigned int num) const
     {
-	throw std::logic_error("not implemented");
+	return Filesystem::isDefault(num);
     }
 
 
     std::pair<bool, unsigned int>
     Btrfs::getDefault() const
     {
-	throw std::logic_error("not implemented");
+	return Filesystem::getDefault();
     }
 
 
     void
     Btrfs::setDefault(unsigned int num) const
     {
-	throw std::logic_error("not implemented");
+	Filesystem::setDefault(num);
+    }
+
+
+    void
+    Btrfs::setDefault(unsigned int num, Plugins::Report& report) const
+    {
+	Filesystem::setDefault(num, report);
+    }
+
+
+    std::pair<bool, unsigned int>
+    Btrfs::getActive() const
+    {
+	return Filesystem::getActive();
     }
 
 
     bool
     Btrfs::isActive(unsigned int num) const
     {
-	throw std::logic_error("not implemented");
+	return Filesystem::isActive(num);
     }
 
 #endif
@@ -1567,7 +1594,7 @@ namespace snapper
 #ifdef HAVE_LIBBTRFS
 	    for (subvolid_t subvolid : deleted_subvolids)
 	    {
-		while (!does_subvolume_exist(subvolume_dir.fd(), subvolid))
+		while (does_subvolume_exist(subvolume_dir.fd(), subvolid))
 		    sleep(1);
 	    }
 #endif
@@ -1587,7 +1614,7 @@ namespace snapper
 	string subvol_option = default_subvolume_name;
 	if (!subvol_option.empty())
 	    subvol_option += "/";
-	subvol_option += ".snapshots";
+	subvol_option += SNAPSHOTS_NAME;
 
 	MntTable mnt_table(root_prefix);
 	mnt_table.parse_fstab();
@@ -1600,7 +1627,7 @@ namespace snapper
 	if (!snapshots)
 	    throw runtime_error("mnt_copy_fs failed");
 
-	mnt_fs_set_target(snapshots, "/.snapshots");
+	mnt_fs_set_target(snapshots, "/" SNAPSHOTS_NAME);
 
 	char* options = mnt_fs_strdup_options(snapshots);
 	mnt_optstr_remove_option(&options, "defaults");
@@ -1619,7 +1646,7 @@ namespace snapper
 	MntTable mnt_table(root_prefix);
 	mnt_table.parse_fstab();
 
-	string mountpoint = (subvolume == "/" ? "" : subvolume) +  "/.snapshots";
+	string mountpoint = (subvolume == "/" ? "" : subvolume) + "/" SNAPSHOTS_NAME;
 	libmnt_fs* snapshots = mnt_table.find_target(mountpoint, MNT_ITER_FORWARD);
 	if (!snapshots)
 	    return;
