@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
- * Copyright (c) 2016 SUSE LLC
+ * Copyright (c) [2016-2024] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -23,8 +23,8 @@
 
 #include "config.h"
 
-#include <string.h>
-#include <errno.h>
+#include <cstring>
+#include <cerrno>
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #include <asm/types.h>
@@ -34,6 +34,11 @@
 #ifdef HAVE_LIBBTRFS
 #include <btrfs/ioctl.h>
 #include <btrfs/send-utils.h>
+#else
+#include <linux/btrfs.h>
+#endif
+#ifdef HAVE_LIBBTRFSUTIL
+#include <btrfsutil.h>
 #endif
 #include <algorithm>
 #include <functional>
@@ -43,45 +48,6 @@
 #include "snapper/BtrfsUtils.h"
 
 
-#ifndef HAVE_LIBBTRFS
-
-#define BTRFS_IOCTL_MAGIC 0x94
-#define BTRFS_PATH_NAME_MAX 4087
-#define BTRFS_SUBVOL_NAME_MAX 4039
-#define BTRFS_SUBVOL_RDONLY (1ULL << 1)
-
-#define BTRFS_IOC_SNAP_CREATE _IOW(BTRFS_IOCTL_MAGIC, 1, struct btrfs_ioctl_vol_args)
-#define BTRFS_IOC_SUBVOL_CREATE _IOW(BTRFS_IOCTL_MAGIC, 14, struct btrfs_ioctl_vol_args)
-#define BTRFS_IOC_SNAP_DESTROY _IOW(BTRFS_IOCTL_MAGIC, 15, struct btrfs_ioctl_vol_args)
-#define BTRFS_IOC_SNAP_CREATE_V2 _IOW(BTRFS_IOCTL_MAGIC, 23, struct btrfs_ioctl_vol_args_v2)
-#define BTRFS_IOC_SYNC _IO(BTRFS_IOCTL_MAGIC, 8)
-
-struct btrfs_ioctl_vol_args
-{
-    __s64 fd;
-    char name[BTRFS_PATH_NAME_MAX + 1];
-};
-
-struct btrfs_ioctl_vol_args_v2
-{
-    __s64 fd;
-    __u64 transid;
-    __u64 flags;
-    __u64 unused[4];
-    char name[BTRFS_SUBVOL_NAME_MAX + 1];
-};
-
-#endif
-
-#ifndef BTRFS_IOC_SUBVOL_GETFLAGS
-#define BTRFS_IOC_SUBVOL_GETFLAGS _IOR(BTRFS_IOCTL_MAGIC, 25, __u64)
-#endif
-
-#ifndef BTRFS_QGROUP_LEVEL_SHIFT
-#define BTRFS_QGROUP_LEVEL_SHIFT 48
-#endif
-
-
 namespace snapper
 {
 
@@ -104,17 +70,60 @@ namespace snapper
 	bool
 	is_subvolume_read_only(int fd)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+	    bool read_only;
+
+	    err = btrfs_util_get_subvolume_read_only_fd(fd, &read_only);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_get_subvolume_read_only_fd() failed", errno);
+
+	    return read_only;
+#else
 	    __u64 flags;
 	    if (ioctl(fd, BTRFS_IOC_SUBVOL_GETFLAGS, &flags) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SUBVOL_GETFLAGS) failed", errno);
 
 	    return flags & BTRFS_SUBVOL_RDONLY;
+#endif
+	}
+
+
+	void
+	set_subvolume_read_only(int fd, bool read_only)
+	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+
+	    err = btrfs_util_set_subvolume_read_only_fd(fd, read_only);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_set_subvolume_read_only_fd() failed", errno);
+#else
+	    __u64 flags;
+	    if (ioctl(fd, BTRFS_IOC_SUBVOL_GETFLAGS, &flags) < 0)
+		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SUBVOL_GETFLAGS) failed", errno);
+
+	    if (read_only)
+		flags |= BTRFS_SUBVOL_RDONLY;
+	    else
+		flags &= ~BTRFS_SUBVOL_RDONLY;
+
+	    if (ioctl(fd, BTRFS_IOC_SUBVOL_SETFLAGS, &flags) < 0)
+		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SUBVOL_SETFLAGS) failed", errno);
+#endif
 	}
 
 
 	void
 	create_subvolume(int fddst, const string& name)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+
+	    err = btrfs_util_create_subvolume_fd(fddst, name.c_str(), 0, NULL, NULL);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_create_subvolume_fd() failed", errno);
+#else
 	    struct btrfs_ioctl_vol_args args;
 	    memset(&args, 0, sizeof(args));
 
@@ -122,12 +131,44 @@ namespace snapper
 
 	    if (ioctl(fddst, BTRFS_IOC_SUBVOL_CREATE, &args) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SUBVOL_CREATE) failed", errno);
+#endif
 	}
 
 
 	void
 	create_snapshot(int fd, int fddst, const string& name, bool read_only, qgroup_t qgroup)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    struct btrfs_util_qgroup_inherit *util_inherit = NULL;
+	    int flags = 0;
+
+	    if (read_only)
+		flags |= BTRFS_UTIL_CREATE_SNAPSHOT_READ_ONLY;
+
+#ifdef ENABLE_BTRFS_QUOTA
+	    size_t size = sizeof(btrfs_qgroup_inherit) + sizeof(((btrfs_qgroup_inherit*) 0)->qgroups[0]);
+	    vector<char> buffer(size, 0);
+
+	    if (qgroup != no_qgroup)
+	    {
+		struct btrfs_qgroup_inherit *inherit;
+
+		inherit = (btrfs_qgroup_inherit*) &buffer[0];
+		inherit->num_qgroups = 1;
+		inherit->num_ref_copies = 0;
+		inherit->num_excl_copies = 0;
+		inherit->qgroups[0] = qgroup;
+		util_inherit = (struct btrfs_util_qgroup_inherit *)inherit;
+	    }
+#endif
+	    enum btrfs_util_error err = btrfs_util_create_snapshot_fd2(fd, fddst, name.c_str(), flags, NULL,
+								       util_inherit);
+	    if (!err)
+		return;
+	    else if (errno != ENOTTY && errno != EINVAL)
+		throw runtime_error_with_errno("btrfs_util_create_snapshot_fd2() failed", errno);
+
+#else
 	    struct btrfs_ioctl_vol_args_v2 args_v2;
 	    memset(&args_v2, 0, sizeof(args_v2));
 
@@ -137,10 +178,11 @@ namespace snapper
 
 #ifdef ENABLE_BTRFS_QUOTA
 
+	    size_t size = sizeof(btrfs_qgroup_inherit) + sizeof(((btrfs_qgroup_inherit*) 0)->qgroups[0]);
+	    vector<char> buffer(size, 0);
+
 	    if (qgroup != no_qgroup)
 	    {
-		size_t size = sizeof(btrfs_qgroup_inherit) + sizeof(((btrfs_qgroup_inherit*) 0)->qgroups[0]);
-		vector<char> buffer(size, 0);
 		struct btrfs_qgroup_inherit* inherit = (btrfs_qgroup_inherit*) &buffer[0];
 
 		inherit->num_qgroups = 1;
@@ -160,6 +202,10 @@ namespace snapper
 	    else if (errno != ENOTTY && errno != EINVAL)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SNAP_CREATE_V2) failed", errno);
 
+#endif
+
+	    // no read-only nor qgroup handling below
+
 	    struct btrfs_ioctl_vol_args args;
 	    memset(&args, 0, sizeof(args));
 
@@ -174,6 +220,13 @@ namespace snapper
 	void
 	delete_subvolume(int fd, const string& name)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+
+	    err = btrfs_util_delete_subvolume_fd(fd, name.c_str(), 0);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_delete_subvolume_fd() failed", errno);
+#else
 	    struct btrfs_ioctl_vol_args args;
 	    memset(&args, 0, sizeof(args));
 
@@ -181,6 +234,7 @@ namespace snapper
 
 	    if (ioctl(fd, BTRFS_IOC_SNAP_DESTROY, &args) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SNAP_DESTROY) failed", errno);
+#endif
 	}
 
 
@@ -189,14 +243,32 @@ namespace snapper
 	void
 	set_default_id(int fd, subvolid_t id)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+
+	    err = btrfs_util_set_default_subvolume_fd(fd, id);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_set_default_subvolume_fd() failed", errno);
+#else
 	    if (ioctl(fd, BTRFS_IOC_DEFAULT_SUBVOL, &id) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_DEFAULT_SUBVOL) failed", errno);
+#endif
 	}
 
 
 	subvolid_t
 	get_default_id(int fd)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+	    uint64_t id;
+
+	    err = btrfs_util_get_default_subvolume_fd(fd, &id);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_get_default_subvolume_fd() failed", errno);
+
+	    return id;
+#else
 	    struct btrfs_ioctl_search_args args;
 	    memset(&args, 0, sizeof(args));
 
@@ -227,12 +299,59 @@ namespace snapper
 		throw std::runtime_error("name != default");
 
 	    return btrfs_disk_key_objectid(&di->location);
+#endif
 	}
 
 
 	string
 	get_subvolume(int fd, subvolid_t id)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+	    char *tmp;
+	    string path;
+
+	    /* This requires CAP_SYS_ADMIN as it uses the TREE_SEARCH ioctl but is fast. */
+	    err = btrfs_util_subvolume_path_fd(fd, id, &tmp);
+	    if (err == BTRFS_UTIL_ERROR_SUBVOLUME_NOT_FOUND)
+		throw runtime_error_with_errno("btrfs_util_subvolume_path_fd() failed", errno);
+
+	    /* Try slower iterative search but without restrictions. */
+	    if (err == BTRFS_UTIL_OK) {
+		path = tmp;
+		free(tmp);
+	    } else if (err == BTRFS_UTIL_ERROR_SEARCH_FAILED || err == BTRFS_UTIL_ERROR_NO_MEMORY) {
+		struct btrfs_util_subvolume_iterator *iter;
+
+		err = btrfs_util_create_subvolume_iterator_fd(fd, 0, 0, &iter);
+		if (err)
+		    throw runtime_error_with_errno("btrfs_util_subvolume_path_fd() failed", errno);
+
+		while (1) {
+		    struct btrfs_util_subvolume_info subvol;
+
+		    err = btrfs_util_subvolume_iterator_next_info(iter, &tmp, &subvol);
+		    if (err != BTRFS_UTIL_OK) {
+			/* Nothing found or other error */
+			btrfs_util_destroy_subvolume_iterator(iter);
+			throw std::runtime_error("get_subvolume() failed");
+		    }
+
+		    if (subvol.id == id) {
+			btrfs_util_destroy_subvolume_iterator(iter);
+			path = tmp;
+			free(tmp);
+			break;
+		    }
+		    free(tmp);
+		}
+	    } else {
+		/* Unknown error */
+		throw std::runtime_error("get_subvolume() failed");
+	    }
+
+	    return path;
+#else
 	    char path[BTRFS_PATH_NAME_MAX + 1];
 
 	    if (btrfs_subvolid_resolve(fd, path, sizeof(path), id) != 0)
@@ -240,13 +359,25 @@ namespace snapper
 
 	    path[BTRFS_PATH_NAME_MAX] = '\0';
 	    return path;
+#endif
 	}
 
 #endif
 
+#ifdef HAVE_LIBBTRFSUTIL
+	subvolid_t
+	get_id(int fd)
+	{
+	    enum btrfs_util_error err;
+	    subvolid_t id;
 
-#ifdef HAVE_LIBBTRFS
+	    err = btrfs_util_subvolume_id_fd(fd, &id);
+	    if (err)
+		throw runtime_error_with_errno("btrfs_util_subvolume_id_fd() failed", errno);
 
+	    return id;
+	}
+#elif defined(HAVE_LIBBTRFS)
 	subvolid_t
 	get_id(int fd)
 	{
@@ -260,11 +391,25 @@ namespace snapper
 
 	    return args.treeid;
 	}
+#endif
 
 
+#ifdef HAVE_LIBBTRFS
 	bool
 	does_subvolume_exist(int fd, subvolid_t subvolid)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+	    struct btrfs_util_subvolume_info subvol_info;
+
+	    err = btrfs_util_subvolume_info_fd(fd, subvolid, &subvol_info);
+	    if (err == BTRFS_UTIL_ERROR_SUBVOLUME_NOT_FOUND)
+		return false;
+	    else if (err)
+		throw runtime_error_with_errno("btrfs_util_subvolume_info_fd() failed", errno);
+
+	    return true;
+#else
 	    struct btrfs_ioctl_search_args args;
 	    struct btrfs_ioctl_search_key* sk = &args.key;
 
@@ -282,7 +427,8 @@ namespace snapper
 	    if (ioctl(fd, BTRFS_IOC_TREE_SEARCH, &args) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_TREE_SEARCH) failed", errno);
 
-	    return sk->nr_items == 0;
+	    return sk->nr_items > 0;
+#endif
 	}
 
 #endif
@@ -320,21 +466,28 @@ namespace snapper
 	    struct btrfs_ioctl_quota_rescan_args args;
 	    memset(&args, 0, sizeof(args));
 
-	    if (ioctl(fd, BTRFS_IOC_QUOTA_RESCAN, &args) < 0)
-		throw runtime_error_with_errno("ioctl(BTRFS_IOC_QUOTA_RESCAN) failed", errno);
-
 	    while (true)
 	    {
-		sleep(1);
-
-		memset(&args, 0, sizeof(args));
-
-		if (ioctl(fd, BTRFS_IOC_QUOTA_RESCAN_STATUS, &args) < 0)
-		    throw runtime_error_with_errno("ioctl(BTRFS_IOC_QUOTA_RESCAN_STATUS) failed", errno);
-
-		if (!args.flags)
+		if (ioctl(fd, BTRFS_IOC_QUOTA_RESCAN, &args) == 0)
 		    break;
+
+		if (errno == EINPROGRESS)
+		{
+		    y2war("waiting for old quota rescan to finish");
+
+		    sleep(1);	// avoid excessive logging (gh#openSUSE/snapper#892)
+
+		    if (ioctl(fd, BTRFS_IOC_QUOTA_RESCAN_WAIT, &args) < 0)
+			throw runtime_error_with_errno("ioctl(BTRFS_IOC_QUOTA_WAIT_RESCAN) failed", errno);
+
+		    continue;
+		}
+
+		throw runtime_error_with_errno("ioctl(BTRFS_IOC_QUOTA_RESCAN) failed", errno);
 	    }
+
+	    if (ioctl(fd, BTRFS_IOC_QUOTA_RESCAN_WAIT, &args) < 0)
+		throw runtime_error_with_errno("ioctl(BTRFS_IOC_QUOTA_WAIT_RESCAN) failed", errno);
 	}
 
 
@@ -448,16 +601,18 @@ namespace snapper
 
 	struct TreeSearchOpts
 	{
-	    TreeSearchOpts(__u32 type) : min_offset(0), max_offset(-1), min_type(type), max_type(type) {}
+	    TreeSearchOpts(__u32 type) : min_type(type), max_type(type) {}
 
-	    __u64 min_offset;
-	    __u64 max_offset;
+	    __u64 min_offset = 0;
+	    __u64 max_offset = -1;
 
 	    __u32 min_type;
 	    __u32 max_type;
 
 	    std::function<void(const struct btrfs_ioctl_search_args& args,
-			       const struct btrfs_ioctl_search_header& sh)> callback;
+			       const struct btrfs_ioctl_search_header& sh)> callback =
+		[](const struct btrfs_ioctl_search_args& args,
+		   const struct btrfs_ioctl_search_header& sh){};
 	};
 
 
@@ -528,6 +683,24 @@ namespace snapper
 	}
 
 
+	bool
+	does_qgroup_exist(int fd, qgroup_t qgroup)
+	{
+	    TreeSearchOpts tree_search_opts(BTRFS_QGROUP_INFO_KEY);
+	    tree_search_opts.min_offset = tree_search_opts.max_offset = qgroup;
+
+	    try
+	    {
+		return qgroups_tree_search(fd, tree_search_opts) > 0;
+	    }
+	    catch (const std::runtime_error& e)
+	    {
+		// happens when quota is disabled
+		return false;
+	    }
+	}
+
+
 	qgroup_t
 	qgroup_find_free(int fd, uint64_t level)
 	{
@@ -560,14 +733,14 @@ namespace snapper
 
 
 	vector<qgroup_t>
-	qgroup_query_children(int fd, qgroup_t parent)
+	qgroup_query_relations(int fd, qgroup_t qgroup)
 	{
 	    vector<qgroup_t> ret;
 
 	    TreeSearchOpts tree_search_opts(BTRFS_QGROUP_RELATION_KEY);
-	    tree_search_opts.min_offset = tree_search_opts.max_offset = parent;
-	    tree_search_opts.callback = [parent, &ret](const struct btrfs_ioctl_search_args& args,
-						       const struct btrfs_ioctl_search_header& sh)
+	    tree_search_opts.min_offset = tree_search_opts.max_offset = qgroup;
+	    tree_search_opts.callback = [&ret](const struct btrfs_ioctl_search_args& args,
+					       const struct btrfs_ioctl_search_header& sh)
 	    {
 		ret.push_back(sh.objectid);
 	    };
@@ -578,6 +751,25 @@ namespace snapper
 	}
 
 
+	vector<qgroup_t>
+	qgroup_query_children(int fd, qgroup_t parent)
+	{
+	    uint64_t level = get_level(parent);
+	    if (level == 0)
+		return {};
+
+	    --level;
+
+	    vector<qgroup_t> ret = qgroup_query_relations(fd, parent);
+
+	    ret.erase(remove_if(ret.begin(), ret.end(),
+		[level](qgroup_t qgroup){ return level != get_level(qgroup); }
+	    ), ret.end());
+
+	    return ret;
+	}
+
+
 	QGroupUsage
 	qgroup_query_usage(int fd, qgroup_t qgroup)
 	{
@@ -585,8 +777,8 @@ namespace snapper
 
 	    TreeSearchOpts tree_search_opts(BTRFS_QGROUP_INFO_KEY);
 	    tree_search_opts.min_offset = tree_search_opts.max_offset = qgroup;
-	    tree_search_opts.callback = [qgroup, &qgroup_usage](const struct btrfs_ioctl_search_args& args,
-								const struct btrfs_ioctl_search_header& sh)
+	    tree_search_opts.callback = [&qgroup_usage](const struct btrfs_ioctl_search_args& args,
+							const struct btrfs_ioctl_search_header& sh)
 	    {
 		struct btrfs_qgroup_info_item info;
 		memcpy(&info, (struct btrfs_qgroup_info_item*)(args.buf + sizeof(sh)), sizeof(info));
@@ -613,8 +805,44 @@ namespace snapper
 	void
 	sync(int fd)
 	{
+#ifdef HAVE_LIBBTRFSUTIL
+	    enum btrfs_util_error err;
+
+	    err = btrfs_util_sync_fd(fd);
+	    if (err)
+		throw runtime_error_with_errno("(btrfs_util_sync_fd() failed", errno);
+#else
 	    if (ioctl(fd, BTRFS_IOC_SYNC) < 0)
 		throw runtime_error_with_errno("ioctl(BTRFS_IOC_SYNC) failed", errno);
+#endif
+	}
+
+
+	Uuid
+	get_uuid(int fd)
+	{
+	    static_assert(BTRFS_UUID_SIZE == 16, "unexpected value of BTRFS_UUID_SIZE");
+
+	    struct btrfs_ioctl_fs_info_args fs_info_args;
+	    if (ioctl(fd, BTRFS_IOC_FS_INFO, &fs_info_args) < 0)
+		throw runtime_error_with_errno("ioctl(BTRFS_IOC_FS_INFO) failed", errno);
+
+	    Uuid uuid;
+	    std::copy(std::begin(fs_info_args.fsid), std::end(fs_info_args.fsid), std::begin(uuid.value));
+	    return uuid;
+	}
+
+
+	Uuid
+	get_uuid(const string& path)
+	{
+	    int fd = open(path.c_str(), O_RDONLY);
+	    if (fd < 0)
+		throw runtime_error_with_errno("open failed", errno);
+
+	    FdCloser fd_closer(fd);
+
+	    return BtrfsUtils::get_uuid(fd);
 	}
 
     }
