@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
- * Copyright (c) [2016,2018] SUSE LLC
+ * Copyright (c) [2016-2024] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -25,13 +25,13 @@
 
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <sys/statvfs.h>
 #include <glob.h>
-#include <string.h>
+#include <cstring>
 #include <mntent.h>
 #include <sys/acl.h>
 #include <acl/libacl.h>
 #include <set>
+#include <regex>
 #include <boost/algorithm/string.hpp>
 
 #include "snapper/Snapper.h"
@@ -45,12 +45,16 @@
 #include "snapper/File.h"
 #include "snapper/AsciiFile.h"
 #include "snapper/Exception.h"
-#include "snapper/Hooks.h"
+#include "snapper/PluginsImpl.h"
+#include "snapper/ComparisonImpl.h"
+#include "snapper/Systemctl.h"
+#include "snapper/Version.h"
+#ifdef ENABLE_BTRFS
 #include "snapper/Btrfs.h"
 #include "snapper/BtrfsUtils.h"
+#endif
 #ifdef ENABLE_SELINUX
 #include "snapper/Selinux.h"
-#include "snapper/Regex.h"
 #endif
 
 
@@ -60,68 +64,70 @@ namespace snapper
 
 
     ConfigInfo::ConfigInfo(const string& config_name, const string& root_prefix)
-	: SysconfigFile(prepend_root_prefix(root_prefix, CONFIGSDIR "/" + config_name)),
+	: SysconfigFile(prepend_root_prefix(root_prefix, CONFIGS_DIR "/" + config_name)),
 	  config_name(config_name), subvolume("/")
     {
-	if (!getValue(KEY_SUBVOLUME, subvolume))
+	if (!get_value(KEY_SUBVOLUME, subvolume))
 	    SN_THROW(InvalidConfigException());
     }
 
 
     void
-    ConfigInfo::checkKey(const string& key) const
+    ConfigInfo::check_key(const string& key) const
     {
 	if (key == KEY_SUBVOLUME || key == KEY_FSTYPE)
 	    SN_THROW(InvalidConfigdataException());
 
 	try
 	{
-	    SysconfigFile::checkKey(key);
+	    SysconfigFile::check_key(key);
 	}
 	catch (const InvalidKeyException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(InvalidConfigdataException());
 	}
     }
 
 
     Snapper::Snapper(const string& config_name, const string& root_prefix, bool disable_filters)
-	: config_info(NULL), filesystem(NULL), snapshots(this), selabel_handle(NULL)
+	: snapshots(this)
     {
 	y2mil("Snapper constructor");
-	y2mil("libsnapper version " VERSION);
-	y2mil("config_name:" << config_name << " disable_filters:" << disable_filters);
-
-#ifdef ENABLE_SELINUX
-	try
-	{
-	    selabel_handle = SelinuxLabelHandle::get_selinux_handle();
-	}
-	catch (const SelinuxException& e)
-	{
-	    SN_RETHROW(e);
-	}
-#endif
+	y2mil("snapper version " VERSION);
+	y2mil("libsnapper version " LIBSNAPPER_VERSION_STRING);
+	y2mil("config_name:" << config_name << " root_prefix:" << root_prefix <<
+	      " disable_filters:" << disable_filters);
 
 	try
 	{
 	    config_info = new ConfigInfo(config_name, root_prefix);
 	}
-	catch (const FileNotFoundException& e)
+	catch (const Exception& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(ConfigNotFoundException());
 	}
 
 	filesystem = Filesystem::create(*config_info, root_prefix);
 
-	// With btrfs backend, it's useless try syncing snapshot RO subvolumes
-	syncSelinuxContexts(filesystem->fstype() == "btrfs");
+#ifdef ENABLE_SELINUX
+	if (_is_selinux_enabled())
+	{
+	    SelinuxLabelHandle* selabel_handle = SelinuxLabelHandle::get_selinux_handle();
+
+	    // With btrfs backend, it's useless try syncing snapshot RO subvolumes
+	    syncSelinuxContexts(selabel_handle, filesystem->fstype() == "btrfs");
+	}
+#endif
 
 	bool sync_acl;
-	if (config_info->getValue(KEY_SYNC_ACL, sync_acl) && sync_acl == true)
+	if (config_info->get_value(KEY_SYNC_ACL, sync_acl) && sync_acl)
 	    syncAcl();
 
-	y2mil("subvolume:" << config_info->getSubvolume() << " filesystem:" <<
+	y2mil("subvolume:" << config_info->get_subvolume() << " filesystem:" <<
 	      filesystem->fstype());
 
 	if (!disable_filters)
@@ -143,6 +149,7 @@ namespace snapper
 	    }
 	    catch (const UmountSnapshotFailedException& e)
 	    {
+		SN_CAUGHT(e);
 	    }
 	}
 
@@ -157,20 +164,40 @@ namespace snapper
     void
     Snapper::loadIgnorePatterns()
     {
-	const list<string> files = glob(FILTERSDIR "/*.txt", GLOB_NOSORT);
-	for (list<string>::const_iterator it = files.begin(); it != files.end(); ++it)
+	const vector<string> etc_files = glob(ETC_FILTERS_DIR "/*.txt", GLOB_NOSORT);
+	const vector<string> usr_files = glob(USR_FILTERS_DIR "/*.txt", GLOB_NOSORT);
+
+	// all files from /etc
+	vector<string> files = etc_files;
+
+	// files from /usr where no corresponding file in /etc exists
+	for (const string& usr_file : usr_files)
+	{
+	    string b1 = basename(usr_file);
+
+	    if (none_of(etc_files.begin(), etc_files.end(), [&b1](const string& etc_file) {
+		string b2 = basename(etc_file);
+		return b1 == b2;
+	    }))
+		files.push_back(usr_file);
+	}
+
+	for (const string& file : files)
 	{
 	    try
 	    {
-		AsciiFileReader asciifile(*it);
+		AsciiFileReader ascii_file_reader(file, Compression::NONE);
 
 		string line;
-		while (asciifile.getline(line))
+		while (ascii_file_reader.read_line(line))
 		    if (!line.empty())
 			ignore_patterns.push_back(line);
+
+		ascii_file_reader.close();
 	    }
-	    catch (const FileNotFoundException& e)
+	    catch (const Exception& e)
 	    {
+		SN_CAUGHT(e);
 	    }
 	}
 
@@ -182,7 +209,7 @@ namespace snapper
     string
     Snapper::subvolumeDir() const
     {
-	return config_info->getSubvolume();
+	return config_info->get_subvolume();
     }
 
 
@@ -210,52 +237,108 @@ namespace snapper
     Snapshots::iterator
     Snapper::createSingleSnapshot(const SCD& scd)
     {
-	return snapshots.createSingleSnapshot(scd);
+	Plugins::Report report;
+	return createSingleSnapshot(scd, report);
     }
 
 
     Snapshots::iterator
     Snapper::createSingleSnapshot(Snapshots::const_iterator parent, const SCD& scd)
     {
-	if (parent == snapshots.end())
-	    SN_THROW(IllegalSnapshotException());
-
-	return snapshots.createSingleSnapshot(parent, scd);
+	Plugins::Report report;
+	return createSingleSnapshot(parent, scd, report);
     }
 
 
     Snapshots::iterator
     Snapper::createSingleSnapshotOfDefault(const SCD& scd)
     {
-	return snapshots.createSingleSnapshotOfDefault(scd);
+	Plugins::Report report;
+	return createSingleSnapshotOfDefault(scd, report);
     }
 
 
     Snapshots::iterator
     Snapper::createPreSnapshot(const SCD& scd)
     {
-	return snapshots.createPreSnapshot(scd);
+	Plugins::Report report;
+	return createPreSnapshot(scd, report);
     }
 
 
     Snapshots::iterator
     Snapper::createPostSnapshot(Snapshots::const_iterator pre, const SCD& scd)
     {
-	return snapshots.createPostSnapshot(pre, scd);
+	Plugins::Report report;
+	return createPostSnapshot(pre, scd, report);
     }
 
 
     void
     Snapper::modifySnapshot(Snapshots::iterator snapshot, const SMD& smd)
     {
-	snapshots.modifySnapshot(snapshot, smd);
+	Plugins::Report report;
+	modifySnapshot(snapshot, smd, report);
     }
 
 
     void
     Snapper::deleteSnapshot(Snapshots::iterator snapshot)
     {
-	snapshots.deleteSnapshot(snapshot);
+	Plugins::Report report;
+	deleteSnapshot(snapshot, report);
+    }
+
+
+    Snapshots::iterator
+    Snapper::createSingleSnapshot(const SCD& scd, Plugins::Report& report)
+    {
+	return snapshots.createSingleSnapshot(scd, report);
+    }
+
+
+    Snapshots::iterator
+    Snapper::createSingleSnapshot(Snapshots::const_iterator parent, const SCD& scd, Plugins::Report& report)
+    {
+	if (parent == snapshots.end())
+	    SN_THROW(IllegalSnapshotException());
+
+	return snapshots.createSingleSnapshot(parent, scd, report);
+    }
+
+
+    Snapshots::iterator
+    Snapper::createSingleSnapshotOfDefault(const SCD& scd, Plugins::Report& report)
+    {
+	return snapshots.createSingleSnapshotOfDefault(scd, report);
+    }
+
+
+    Snapshots::iterator
+    Snapper::createPreSnapshot(const SCD& scd, Plugins::Report& report)
+    {
+	return snapshots.createPreSnapshot(scd, report);
+    }
+
+
+    Snapshots::iterator
+    Snapper::createPostSnapshot(Snapshots::const_iterator pre, const SCD& scd, Plugins::Report& report)
+    {
+	return snapshots.createPostSnapshot(pre, scd, report);
+    }
+
+
+    void
+    Snapper::modifySnapshot(Snapshots::iterator snapshot, const SMD& smd, Plugins::Report& report)
+    {
+	snapshots.modifySnapshot(snapshot, smd, report);
+    }
+
+
+    void
+    Snapper::deleteSnapshot(Snapshots::iterator snapshot, Plugins::Report& report)
+    {
+	snapshots.deleteSnapshot(snapshot, report);
     }
 
 
@@ -276,29 +359,29 @@ namespace snapper
 
 	try
 	{
-	    SysconfigFile sysconfig(prepend_root_prefix(root_prefix, SYSCONFIGFILE));
+	    SysconfigFile sysconfig(prepend_root_prefix(root_prefix, SYSCONFIG_FILE));
 	    vector<string> config_names;
-	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.get_value("SNAPPER_CONFIGS", config_names);
 
-	    for (vector<string>::const_iterator it = config_names.begin(); it != config_names.end(); ++it)
+	    for (const string& config_name : config_names)
 	    {
 		try
 		{
-		    config_infos.push_back(getConfig(*it, root_prefix));
+		    config_infos.push_back(getConfig(config_name, root_prefix));
 		}
-		catch (const FileNotFoundException& e)
+		catch (const Exception& e)
 		{
-		    y2err("config '" << *it << "' not found");
-		}
-		catch (const InvalidConfigException& e)
-		{
-		    y2err("config '" << *it << "' is invalid");
+		    SN_CAUGHT(e);
+
+		    y2err("reading config '" << config_name << "' failed");
 		}
 	    }
 	}
-	catch (const FileNotFoundException& e)
+	catch (const Exception& e)
 	{
-	    SN_THROW(ListConfigsFailedException("sysconfig-file not found"));
+	    SN_CAUGHT(e);
+
+	    SN_THROW(ListConfigsFailedException("reading sysconfig-file failed"));
 	}
 
 	return config_infos;
@@ -309,6 +392,16 @@ namespace snapper
     Snapper::createConfig(const string& config_name, const string& root_prefix,
 			  const string& subvolume, const string& fstype,
 			  const string& template_name)
+    {
+	Plugins::Report report;
+	createConfig(config_name, root_prefix, subvolume, fstype, template_name, report);
+    }
+
+
+    void
+    Snapper::createConfig(const string& config_name, const string& root_prefix,
+			  const string& subvolume, const string& fstype,
+			  const string& template_name, Plugins::Report& report)
     {
 	y2mil("Snapper create-config");
 	y2mil("libsnapper version " VERSION);
@@ -328,13 +421,19 @@ namespace snapper
 	list<ConfigInfo> configs = getConfigs(root_prefix);
 	for (list<ConfigInfo>::const_iterator it = configs.begin(); it != configs.end(); ++it)
 	{
-	    if (it->getSubvolume() == subvolume)
+	    if (it->get_subvolume() == subvolume)
 	    {
 		SN_THROW(CreateConfigFailedException("subvolume already covered"));
 	    }
 	}
 
-	if (access(string(CONFIGTEMPLATEDIR "/" + template_name).c_str(), R_OK) != 0)
+	string template_file;
+
+	try
+	{
+	    template_file = locate_file(template_name, ETC_CONFIG_TEMPLATE_DIR, USR_CONFIG_TEMPLATE_DIR);
+	}
+	catch (...)
 	{
 	    SN_THROW(CreateConfigFailedException("cannot access template config"));
 	}
@@ -346,42 +445,60 @@ namespace snapper
 	}
 	catch (const InvalidConfigException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(CreateConfigFailedException("invalid filesystem type"));
 	}
 	catch (const ProgramNotInstalledException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(CreateConfigFailedException(e.what()));
 	}
 
+	Plugins::create_config(Plugins::Stage::PRE_ACTION, subvolume, filesystem.get(), report);
+
 	try
 	{
-	    SysconfigFile sysconfig(SYSCONFIGFILE);
+	    SysconfigFile sysconfig(SYSCONFIG_FILE);
 	    vector<string> config_names;
-	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.get_value("SNAPPER_CONFIGS", config_names);
 	    if (find(config_names.begin(), config_names.end(), config_name) != config_names.end())
 	    {
 		SN_THROW(CreateConfigFailedException("config already exists"));
 	    }
 
 	    config_names.push_back(config_name);
-	    sysconfig.setValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.set_value("SNAPPER_CONFIGS", config_names);
+
+	    sysconfig.save();
 	}
-	catch (const FileNotFoundException& e)
+	catch (const Exception& e)
 	{
-	    SN_THROW(CreateConfigFailedException("sysconfig-file not found"));
+	    SN_CAUGHT(e);
+
+	    SN_THROW(CreateConfigFailedException(e.what()));
 	}
 
+	bool timeline_create = false;
+
 	try
 	{
-	    SysconfigFile config(CONFIGTEMPLATEDIR "/" + template_name);
+	    SysconfigFile config(template_file);
 
-	    config.setName(CONFIGSDIR "/" + config_name);
+	    config.set_name(CONFIGS_DIR "/" + config_name);
 
-	    config.setValue(KEY_SUBVOLUME, subvolume);
-	    config.setValue(KEY_FSTYPE, filesystem->fstype());
+	    config.set_value(KEY_SUBVOLUME, subvolume);
+	    config.set_value(KEY_FSTYPE, filesystem->fstype());
+
+	    config.save();
+
+	    config.get_value(KEY_TIMELINE_CREATE, timeline_create);
 	}
-	catch (const FileNotFoundException& e)
+	catch (const Exception& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(CreateConfigFailedException("modifying config failed"));
 	}
 
@@ -393,46 +510,68 @@ namespace snapper
 	{
 	    SN_CAUGHT(e);
 
-	    SysconfigFile sysconfig(SYSCONFIGFILE);
+	    SysconfigFile sysconfig(SYSCONFIG_FILE);
 	    vector<string> config_names;
-	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.get_value("SNAPPER_CONFIGS", config_names);
 	    config_names.erase(remove(config_names.begin(), config_names.end(), config_name),
 			       config_names.end());
-	    sysconfig.setValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.set_value("SNAPPER_CONFIGS", config_names);
 
-	    SystemCmd cmd(RMBIN " " + quote(CONFIGSDIR "/" + config_name));
+	    sysconfig.save();
+
+	    SystemCmd cmd({ RM_BIN, "--", CONFIGS_DIR "/" + config_name });
 
 	    SN_RETHROW(e);
 	}
 
-	Hooks::create_config(subvolume, filesystem.get());
+	if (timeline_create)
+	{
+	    systemctl_enable_timeline(true);
+	}
+
+	Plugins::create_config(Plugins::Stage::POST_ACTION, subvolume, filesystem.get(), report);
     }
 
 
     void
     Snapper::deleteConfig(const string& config_name, const string& root_prefix)
+    {
+	Plugins::Report report;
+	deleteConfig(config_name, root_prefix, report);
+    }
+
+
+    void
+    Snapper::deleteConfig(const string& config_name, const string& root_prefix, Plugins::Report& report)
     {
 	y2mil("Snapper delete-config");
 	y2mil("libsnapper version " VERSION);
 
 	unique_ptr<Snapper> snapper(new Snapper(config_name, root_prefix));
 
-	Hooks::delete_config(snapper->subvolumeDir(), snapper->getFilesystem());
+	Plugins::delete_config(Plugins::Stage::PRE_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+			       report);
 
 	Snapshots& snapshots = snapper->getSnapshots();
+
+	Snapshots::const_iterator default_snapshot = snapshots.getDefault();
+	Snapshots::const_iterator active_snapshot = snapshots.getActive();
+
 	for (Snapshots::iterator it = snapshots.begin(); it != snapshots.end(); )
 	{
 	    Snapshots::iterator tmp = it++;
 
-	    if (tmp->isCurrent())
+	    if (tmp->isCurrent() || tmp == default_snapshot || tmp == active_snapshot)
 		continue;
 
 	    try
 	    {
-		snapper->deleteSnapshot(tmp);
+		snapper->deleteSnapshot(tmp, report);
 	    }
 	    catch (const DeleteSnapshotFailedException& e)
 	    {
+		SN_CAUGHT(e);
+
 		// ignore, Filesystem->deleteConfig will fail anyway
 	    }
 	}
@@ -443,10 +582,12 @@ namespace snapper
 	}
 	catch (const DeleteConfigFailedException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SN_THROW(DeleteConfigFailedException("deleting snapshot failed"));
 	}
 
-	SystemCmd cmd1(RMBIN " " + quote(CONFIGSDIR "/" + config_name));
+	SystemCmd cmd1({ RM_BIN, "--", CONFIGS_DIR "/" + config_name });
 	if (cmd1.retcode() != 0)
 	{
 	    SN_THROW(DeleteConfigFailedException("deleting config-file failed"));
@@ -454,17 +595,26 @@ namespace snapper
 
 	try
 	{
-	    SysconfigFile sysconfig(SYSCONFIGFILE);
+	    SysconfigFile sysconfig(SYSCONFIG_FILE);
 	    vector<string> config_names;
-	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.get_value("SNAPPER_CONFIGS", config_names);
 	    config_names.erase(remove(config_names.begin(), config_names.end(), config_name),
 			       config_names.end());
-	    sysconfig.setValue("SNAPPER_CONFIGS", config_names);
+	    sysconfig.set_value("SNAPPER_CONFIGS", config_names);
+
+	    sysconfig.save();
 	}
-	catch (const FileNotFoundException& e)
+	catch (const Exception& e)
 	{
-	    SN_THROW(DeleteConfigFailedException("sysconfig-file not found"));
+	    SN_CAUGHT(e);
+
+	    SN_THROW(DeleteConfigFailedException("modifying sysconfig-file failed"));
 	}
+
+	// TODO potentially disable snapper-timeline.timer
+
+	Plugins::delete_config(Plugins::Stage::POST_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+			       report);
     }
 
 
@@ -472,7 +622,7 @@ namespace snapper
     Snapper::setConfigInfo(const map<string, string>& raw)
     {
 	for (map<string, string>::const_iterator it = raw.begin(); it != raw.end(); ++it)
-	    config_info->setValue(it->first, it->second);
+	    config_info->set_value(it->first, it->second);
 
 	config_info->save();
 
@@ -482,9 +632,25 @@ namespace snapper
 	    raw.find(KEY_SYNC_ACL) != raw.end())
 	{
 	    bool sync_acl;
-	    if (config_info->getValue(KEY_SYNC_ACL, sync_acl) && sync_acl == true)
+	    if (config_info->get_value(KEY_SYNC_ACL, sync_acl) && sync_acl)
 		syncAcl();
 	}
+
+	if (raw.find(KEY_TIMELINE_CREATE) != raw.end())
+	{
+	    bool timeline_create;
+	    if (config_info->get_value(KEY_TIMELINE_CREATE, timeline_create))
+	    {
+		if (timeline_create)
+		{
+		    systemctl_enable_timeline(true);
+		}
+		else
+		{
+		    // TODO potentially disable snapper-timeline.timer
+		}
+	    }
+	}
     }
 
 
@@ -493,26 +659,26 @@ namespace snapper
     {
 	vector<uid_t> uids;
 	vector<string> users;
-	if (config_info->getValue(KEY_ALLOW_USERS, users))
+	if (config_info->get_value(KEY_ALLOW_USERS, users))
 	{
-	    for (vector<string>::const_iterator it = users.begin(); it != users.end(); ++it)
+	    for (const string& user : users)
 	    {
 		uid_t uid;
-		if (!get_user_uid(it->c_str(), uid))
-		    SN_THROW(InvalidUserException());
+		if (!get_user_uid(user.c_str(), uid))
+		    SN_THROW(InvalidUserException(user));
 		uids.push_back(uid);
 	    }
 	}
 
 	vector<gid_t> gids;
 	vector<string> groups;
-	if (config_info->getValue(KEY_ALLOW_GROUPS, groups))
+	if (config_info->get_value(KEY_ALLOW_GROUPS, groups))
 	{
-	    for (vector<string>::const_iterator it = groups.begin(); it != groups.end(); ++it)
+	    for (const string& group : groups)
 	    {
 		gid_t gid;
-		if (!get_group_gid(it->c_str(), gid))
-		    SN_THROW(InvalidGroupException());
+		if (!get_group_gid(group.c_str(), gid))
+		    SN_THROW(InvalidGroupException(group));
 		gids.push_back(gid);
 	    }
 	}
@@ -741,6 +907,36 @@ namespace snapper
     }
 
 
+    void
+    Snapper::calculateUsedSpace() const
+    {
+#ifdef ENABLE_BTRFS_QUOTA
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(QuotaException("quota only supported with btrfs"));
+
+	try
+	{
+	    SDir general_dir = btrfs->openGeneralDir();
+
+	    quota_rescan(general_dir.fd());
+	    sync(general_dir.fd());
+	}
+	catch (...)
+	{
+	    SN_THROW(QuotaException("quota rescan or sync failed"));
+	}
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
     QuotaData
     Snapper::queryQuotaData() const
     {
@@ -758,16 +954,19 @@ namespace snapper
 	// Tests have shown that without a rescan and sync here the quota data
 	// is incorrect.
 
-	quota_rescan(general_dir.fd());
-	sync(general_dir.fd());
-
-	struct statvfs64 fsbuf;
-	if (fstatvfs64(general_dir.fd(), &fsbuf) != 0)
-	    SN_THROW(QuotaException("statvfs64 failed"));
+	try
+	{
+	    quota_rescan(general_dir.fd());
+	    sync(general_dir.fd());
+	}
+	catch (...)
+	{
+	    SN_THROW(QuotaException("quota rescan or sync failed"));
+	}
 
 	QuotaData quota_data;
 
-	quota_data.size = fsbuf.f_blocks * fsbuf.f_bsize;
+	std::tie(quota_data.size, std::ignore) = general_dir.statvfs();
 
 	QGroupUsage qgroup_usage = qgroup_query_usage(general_dir.fd(), btrfs->getQGroup());
 	quota_data.used = qgroup_usage.exclusive;
@@ -788,25 +987,64 @@ namespace snapper
     }
 
 
+    FreeSpaceData
+    Snapper::queryFreeSpaceData() const
+    {
+#ifdef ENABLE_BTRFS
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(FreeSpaceException("free space only supported with btrfs"));
+
+	SDir general_dir = btrfs->openGeneralDir();
+
+	try
+	{
+	    filesystem->sync();
+	}
+	catch (...)
+	{
+	    SN_THROW(FreeSpaceException("filesystem sync failed"));
+	}
+
+	FreeSpaceData free_space_data;
+	std::tie(free_space_data.size, free_space_data.free) = general_dir.statvfs();
+
+	y2mil("size:" << free_space_data.size << " free:" << free_space_data.free);
+
+	if (free_space_data.free > free_space_data.size)
+	    SN_THROW(FreeSpaceException("impossible free space values"));
+
+	return free_space_data;
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
     void
-    Snapper::syncSelinuxContexts(bool skip_snapshot_dir) const
+    Snapper::syncSelinuxContexts(SelinuxLabelHandle* selabel_handle, bool skip_snapshot_dir) const
     {
 #ifdef ENABLE_SELINUX
 	try
 	{
 	    SDir subvol_dir = openSubvolumeDir();
-	    SDir infos_dir(subvol_dir, ".snapshots");
+	    SDir infos_dir(subvol_dir, SNAPSHOTS_NAME);
 
 	    if (infos_dir.restorecon(selabel_handle))
 	    {
-		syncSelinuxContextsInInfosDir(skip_snapshot_dir);
+		syncSelinuxContextsInInfosDir(selabel_handle, skip_snapshot_dir);
 	    }
 	    else
 	    {
 		SnapperContexts scons;
 
 		if (infos_dir.fsetfilecon(scons.subvolume_context()))
-		    syncSelinuxContextsInInfosDir(skip_snapshot_dir);
+		    syncSelinuxContextsInInfosDir(selabel_handle, skip_snapshot_dir);
 	    }
 	}
 	catch (const SelinuxException& e)
@@ -819,11 +1057,10 @@ namespace snapper
 
 
     void
-    Snapper::syncSelinuxContextsInInfosDir(bool skip_snapshot_dir) const
+    Snapper::syncSelinuxContextsInInfosDir(SelinuxLabelHandle* selabel_handle, bool skip_snapshot_dir) const
     {
 #ifdef ENABLE_SELINUX
-	Regex rx("^[0-9]+$");
-	Regex rx_filelist("^filelist-[0-9]+.txt$");
+	static const regex rx("[0-9]+", regex::extended);
 
 	y2deb("Syncing Selinux contexts in infos dir");
 
@@ -832,7 +1069,7 @@ namespace snapper
 	vector<string> infos = infos_dir.entries();
 	for (vector<string>::const_iterator it1 = infos.begin(); it1 != infos.end(); ++it1)
 	{
-	    if (!rx.match(*it1))
+	    if (!regex_match(*it1, rx))
 		continue;
 
 	    SDir info_dir(infos_dir, *it1);
@@ -843,16 +1080,13 @@ namespace snapper
 
 	    if (!skip_snapshot_dir)
 	    {
-		SFile snapshot_dir(info_dir, "snapshot");
+		SFile snapshot_dir(info_dir, SNAPSHOT_NAME);
 		snapshot_dir.restorecon(selabel_handle);
 	    }
 
-	    vector<string> info_content = info_dir.entries();
+	    vector<string> info_content = info_dir.entries(is_filelist_file);
 	    for (vector<string>::const_iterator it2 = info_content.begin(); it2 != info_content.end(); ++it2)
 	    {
-		if (!rx_filelist.match(*it2))
-		    continue;
-
 		SFile fl(info_dir, *it2);
 		fl.restorecon(selabel_handle);
 	    }
@@ -919,6 +1153,30 @@ namespace snapper
     }
 
 
+    Compression
+    Snapper::get_compression() const
+    {
+	Compression compression = Compression::GZIP;
+
+	string tmp;
+
+	if (config_info->get_value(KEY_COMPRESSION, tmp))
+	{
+	    if (tmp == "none")
+		compression = Compression::NONE;
+	    else if (tmp == "gzip")
+		compression = Compression::GZIP;
+	    else if (tmp == "zstd")
+		compression = Compression::ZSTD;
+	}
+
+	if (!is_available(compression))
+	    compression = Compression::NONE;
+
+	return compression;
+    }
+
+
     const char*
     Snapper::compileVersion()
     {
@@ -936,6 +1194,11 @@ namespace snapper
 #endif
 	    "btrfs,"
 
+#ifndef ENABLE_BCACHEFS
+	    "no-"
+#endif
+	    "bcachefs,"
+
 #ifndef ENABLE_LVM
 	    "no-"
 #endif
